/*********** Generated on 2026/07/26 21:16:53 by reswrap version 6.0.0 *********/

/* Created by reswrap from file cursor_hand.gif */
extern const unsigned char cursor_hand_gif[];

/* Created by reswrap from file about.png */
extern const unsigned char about_png[];

/* Created by reswrap from file gogglesmm_16.png */
extern const unsigned char gogglesmm_16_png[];

/* Created by reswrap from file gogglesmm_32.png */
extern const unsigned char gogglesmm_32_png[];

/* Created by reswrap from file x16_accessories_text_editor.png */
extern const unsigned char x16_accessories_text_editor_png[];

/* Created by reswrap from file x16_application_rss_xml.png */
extern const unsigned char x16_application_rss_xml_png[];

/* Created by reswrap from file x16_applications_internet.png */
extern const unsigned char x16_applications_internet_png[];

/* Created by reswrap from file x16_audio_volume_high.png */
extern const unsigned char x16_audio_volume_high_png[];

/* Created by reswrap from file x16_audio_volume_low.png */
extern const unsigned char x16_audio_volume_low_png[];

/* Created by reswrap from file x16_audio_volume_medium.png */
extern const unsigned char x16_audio_volume_medium_png[];

/* Created by reswrap from file x16_audio_volume_muted.png */
extern const unsigned char x16_audio_volume_muted_png[];

/* Created by reswrap from file x16_audio_x_generic.png */
extern const unsigned char x16_audio_x_generic_png[];

/* Created by reswrap from file x16_bookmark_new.png */
extern const unsigned char x16_bookmark_new_png[];

/* Created by reswrap from file x16_document_open.png */
extern const unsigned char x16_document_open_png[];

/* Created by reswrap from file x16_document_save.png */
extern const unsigned char x16_document_save_png[];

/* Created by reswrap from file x16_edit_copy.png */
extern const unsigned char x16_edit_copy_png[];

/* Created by reswrap from file x16_edit_cut.png */
extern const unsigned char x16_edit_cut_png[];

/* Created by reswrap from file x16_edit_delete.png */
extern const unsigned char x16_edit_delete_png[];

/* Created by reswrap from file x16_edit_find.png */
extern const unsigned char x16_edit_find_png[];

/* Created by reswrap from file x16_edit_paste.png */
extern const unsigned char x16_edit_paste_png[];

/* Created by reswrap from file x16_edit_undo.png */
extern const unsigned char x16_edit_undo_png[];

/* Created by reswrap from file x16_exit.png */
extern const unsigned char x16_exit_png[];

/* Created by reswrap from file x16_folder.png */
extern const unsigned char x16_folder_png[];

/* Created by reswrap from file x16_folder_open.png */
extern const unsigned char x16_folder_open_png[];

/* Created by reswrap from file x16_go_bottom.png */
extern const unsigned char x16_go_bottom_png[];

/* Created by reswrap from file x16_go_home.png */
extern const unsigned char x16_go_home_png[];

/* Created by reswrap from file x16_help_browser.png */
extern const unsigned char x16_help_browser_png[];

/* Created by reswrap from file x16_image_x_generic.png */
extern const unsigned char x16_image_x_generic_png[];

/* Created by reswrap from file x16_list_add.png */
extern const unsigned char x16_list_add_png[];

/* Created by reswrap from file x16_list_remove.png */
extern const unsigned char x16_list_remove_png[];

/* Created by reswrap from file x16_media_optical.png */
extern const unsigned char x16_media_optical_png[];

/* Created by reswrap from file x16_media_playback_pause.png */
extern const unsigned char x16_media_playback_pause_png[];

/* Created by reswrap from file x16_media_playback_start.png */
extern const unsigned char x16_media_playback_start_png[];

/* Created by reswrap from file x16_media_playback_stop.png */
extern const unsigned char x16_media_playback_stop_png[];

/* Created by reswrap from file x16_media_skip_backward.png */
extern const unsigned char x16_media_skip_backward_png[];

/* Created by reswrap from file x16_media_skip_forward.png */
extern const unsigned char x16_media_skip_forward_png[];

/* Created by reswrap from file x16_preferences_desktop.png */
extern const unsigned char x16_preferences_desktop_png[];

/* Created by reswrap from file x16_process_working.png */
extern const unsigned char x16_process_working_png[];

/* Created by reswrap from file x16_stock_attach.png */
extern const unsigned char x16_stock_attach_png[];

/* Created by reswrap from file x16_status_error.png */
extern const unsigned char x16_status_error_png[];

/* Created by reswrap from file x16_system_users.png */
extern const unsigned char x16_system_users_png[];

/* Created by reswrap from file x16_text_x_generic.png */
extern const unsigned char x16_text_x_generic_png[];

/* Created by reswrap from file x16_view_refresh.png */
extern const unsigned char x16_view_refresh_png[];

/* Created by reswrap from file x16_view_sort_descending.png */
extern const unsigned char x16_view_sort_descending_png[];

/* Created by reswrap from file x16_window_close.png */
extern const unsigned char x16_window_close_png[];

/* Created by reswrap from file x16_x_office_presentation.png */
extern const unsigned char x16_x_office_presentation_png[];

/* Created by reswrap from file x22_applications_internet.png */
extern const unsigned char x22_applications_internet_png[];

/* Created by reswrap from file x22_applications_multimedia.png */
extern const unsigned char x22_applications_multimedia_png[];

/* Created by reswrap from file x22_applications_rss_xml.png */
extern const unsigned char x22_applications_rss_xml_png[];

/* Created by reswrap from file x22_audio_volume_high.png */
extern const unsigned char x22_audio_volume_high_png[];

/* Created by reswrap from file x22_audio_volume_low.png */
extern const unsigned char x22_audio_volume_low_png[];

/* Created by reswrap from file x22_audio_volume_medium.png */
extern const unsigned char x22_audio_volume_medium_png[];

/* Created by reswrap from file x22_audio_volume_muted.png */
extern const unsigned char x22_audio_volume_muted_png[];

/* Created by reswrap from file x22_audio_x_generic.png */
extern const unsigned char x22_audio_x_generic_png[];

/* Created by reswrap from file x22_document_open.png */
extern const unsigned char x22_document_open_png[];

/* Created by reswrap from file x22_document_properties.png */
extern const unsigned char x22_document_properties_png[];

/* Created by reswrap from file x22_drive_harddisk.png */
extern const unsigned char x22_drive_harddisk_png[];

/* Created by reswrap from file x22_folder.png */
extern const unsigned char x22_folder_png[];

/* Created by reswrap from file x22_image_x_generic.png */
extern const unsigned char x22_image_x_generic_png[];

/* Created by reswrap from file x22_media_playback_pause.png */
extern const unsigned char x22_media_playback_pause_png[];

/* Created by reswrap from file x22_media_playback_start.png */
extern const unsigned char x22_media_playback_start_png[];

/* Created by reswrap from file x22_media_playback_stop.png */
extern const unsigned char x22_media_playback_stop_png[];

/* Created by reswrap from file x22_media_skip_backward.png */
extern const unsigned char x22_media_skip_backward_png[];

/* Created by reswrap from file x22_media_skip_forward.png */
extern const unsigned char x22_media_skip_forward_png[];

/* Created by reswrap from file x22_preferences_system.png */
extern const unsigned char x22_preferences_system_png[];

/* Created by reswrap from file x22_text_x_generic.png */
extern const unsigned char x22_text_x_generic_png[];

/* Created by reswrap from file x22_user_bookmarks.png */
extern const unsigned char x22_user_bookmarks_png[];

/* Created by reswrap from file x22_user_home.png */
extern const unsigned char x22_user_home_png[];

/* Created by reswrap from file x22_x_office_presentation.png */
extern const unsigned char x22_x_office_presentation_png[];

/* Created by reswrap from file x128_media_optical.png */
extern const unsigned char x128_media_optical_png[];

